-- Lua optimizations
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strjoin = strjoin
local strlen = strlen
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tonumber = tonumber
local wipe = wipe
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"
local FONTEND = FONT_COLOR_CODE_CLOSE
local NORMALFONT = NORMAL_FONT_COLOR_CODE
local YELLOWFONT = LIGHTYELLOW_FONT_COLOR_CODE
local GRAYFONT = GRAY_FONT_COLOR_CODE
local DUNGEON_DIFFICULTY1 = DUNGEON_DIFFICULTY1 -- 5 man
local DUNGEON_DIFFICULTY2 = DUNGEON_DIFFICULTY2 -- 5 man (Heroic)
local RAID_DIFFICULTY1 = RAID_DIFFICULTY1 or "10 man" -- pre-3.2, 
local RAID_DIFFICULTY2 = RAID_DIFFICULTY2 or "25 man"
local RAID_DIFFICULTY3 = RAID_DIFFICULTY3 or "10 man (Heroic)"
local RAID_DIFFICULTY4 = RAID_DIFFICULTY4 or "25 man (Heroic)"
local EXPANSION_NAME0 = EXPANSION_NAME0 -- "Classic"
local EXPANSION_NAME1 = EXPANSION_NAME1 -- "The Burning Crusade"
local EXPANSION_NAME2 = EXPANSION_NAME2 -- "Wrath of the Lich King"
local LFG_TYPE_DUNGEON = LFG_TYPE_DUNGEON -- "Dungeon"
local LFG_TYPE_RAID = LFG_TYPE_RAID -- "Raid"
local EMBLEM_SYMBOL = EMBLEM_SYMBOL -- "Icon"
local COLOR = COLOR -- "Color"
local DEFAULTS = DEFAULTS -- "Defaults"
local RESET_TO_DEFAULT = RESET_TO_DEFAULT -- "Reset to Default"
local DELETE = DELETE -- "DELETE"

local Indicators = {
	ICON_STAR = ICON_LIST[1] .. "16:16:0:0|t",
	ICON_CIRCLE = ICON_LIST[2] .. "16:16:0:0|t",
	ICON_DIAMOND = ICON_LIST[3] .. "16:16:0:0|t",
	ICON_TRIANGLE = ICON_LIST[4] .. "16:16:0:0|t",
	ICON_MOON = ICON_LIST[5] .. "16:16:0:0|t",
	ICON_SQUARE = ICON_LIST[6] .. "16:16:0:0|t",
	ICON_CROSS = ICON_LIST[7] .. "16:16:0:0|t",
	ICON_SKULL = ICON_LIST[8] .. "16:16:0:0|t",
	BLANK = "None",
}

local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip, indicatortip

local thisToon = UnitName("player") .. " - " .. GetRealmName()
local selectedInstance -- for re-ordering the instance list
local selectedCategory

local db
local defaultDB = {
	DBVersion = 6,
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
	Indicators = {
		D1Indicator = "BLANK", -- indicator: ICON_*, BLANK
		D1Text = "N",
		D1Color = { 0, 0.6, 0, 1, }, -- dark green
		D2Indicator = "BLANK", -- indicator
		D2Text = "H",
		D2Color = { 0, 1, 0, 1, }, -- green
		R1Indicator = "BLANK", -- indicator: ICON_*, BLANK
		R1Text = "N",
		R1Color = { 0.6, 0.6, 0, 1, }, -- dark yellow
		R2Indicator = "BLANK", -- indicator
		R2Text = "H",
		R2Color = { 0.6, 0, 0, 1, }, -- dark red
		R3Indicator = "BLANK", -- indicator: ICON_*, BLANK
		R3Text = "N+",
		R3Color = { 1, 1, 0, 1, }, -- yellow
		R4Indicator = "BLANK", -- indicator
		R4Text = "H+",
		R4Color = { 1, 0, 0, 1, }, -- red
	},
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
		NewInstanceShow = false,
		ShowCategories = true,
		ShowSoloCategory = false,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- Raid: boolean
					-- Expansion: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	Categories = {
		D0 = { Name = EXPANSION_NAME0 .. ": " .. LFG_TYPE_DUNGEON, Order = 1 },
		R0 = { Name = EXPANSION_NAME0 .. ": " .. LFG_TYPE_RAID, Order = 2 },
		D1 = { Name = EXPANSION_NAME1 .. ": " .. LFG_TYPE_DUNGEON, Order = 3 },
		R1 = { Name = EXPANSION_NAME1 .. ": " .. LFG_TYPE_RAID, Order = 4 },
		D2 = { Name = EXPANSION_NAME2 .. ": " .. LFG_TYPE_DUNGEON, Order = 5 },
		R2 = { Name = EXPANSION_NAME2 .. ": " .. LFG_TYPE_RAID, Order = 6 },
		ZZ = { Name = L["Uncategorised"], Order = 7 },
	},
	MinimapIcon = { },
}

-- general helper functions below

local function ColorCodeOpen(color)
	return format("|c%02x%02x%02x%02x", math.floor(color[4] * 255), math.floor(color[1] * 255), math.floor(color[2] * 255), math.floor(color[3] * 255))
end

local function ClassColorise(class, targetstring)
	local RAID_CLASS_COLORS = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	local color = {
		RAID_CLASS_COLORS[class].r,
		RAID_CLASS_COLORS[class].g,
		RAID_CLASS_COLORS[class].b,
		1,
	}
	return ColorCodeOpen(color) .. targetstring .. FONTEND
end

local function TableLen(table)
	local i = 0
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end

local function InstanceCategory(instance)
	if not instance then return nil end
	local instance = SavedInstances.db.Instances[instance]
	if (instance.Raid == nil) or (instance.Expansion == nil) then
		return "ZZ"
	else
		return ((instance.Raid and "R") or (not instance.Raid and "D")) .. tostring(instance.Expansion)
	end
end

local function OrderedInstances(targetcategory)
	-- TODO: fix this so it actually does things in order
	if not targetcategory then return { } end
	local list = { }
	for instance, a in pairs(SavedInstances.db.Instances) do
		if targetcategory == InstanceCategory(instance) then
			list[#list + 1] = instance
		end
	end
	return list
end

local function CategorySize(category)
	if not category then return nil end
	local i = 0
	for instance, a in pairs(SavedInstances.db.Instances) do
		if category == InstanceCategory(instance) then
			i = i + 1
		end
	end
	return i
end

local function OrderedCategories()
	local list = { }
	for category, a in pairs(SavedInstances.db.Categories) do
		list[a.Order] = category
	end
	return list
end

-- local addon functions below

local function DifficultyString(instance, diff)
	local instance = SavedInstances.db.Instances[instance]
	local prefs = SavedInstances.db.Indicators
	local setting = ((instance.Raid and "R") or (not instance.Raid and "D")) .. tostring(instance.Expansion)
	local head, tail = strsplit("ICON", prefs[setting.."Text"])
	head = ColorCodeOpen(prefs[setting.."Color"]) .. head .. FONTEND
	tail = ColorCodeOpen(prefs[setting.."Color"]) .. tail .. FONTEND
	
	if prefs[setting.."Indicator"] == "BLANK" then
		return head .. 	gsub(prefs[setting.."Indicator"], "ICON", "") .. tail
	elseif prefs[setting.."Indicator"] == "BLANK" then
		return head .. 	gsub(prefs[setting.."Indicator"], "ICON", Indicators[prefs[setting.."Indicator"]]) .. tail
	end
end

local function CullStaleInstances()
	for instance, a in pairs(SavedInstances.db.Instances) do
		for toon, b in pairs(a) do
			if type(b) == "table" then
				for difficulty, c in pairs(b) do
					if c.Expires < time() then
						SavedInstances.db.Instances[instance][toon][difficulty] = nil
					end
				end
				if TableLen(b) == 0 then
					SavedInstances.db.Instances[instance][toon] = nil
				end
			end
		end
	end
end

local function HideIndicatorTooltip()
	QTip:Release(indicatortip)
	indicatortip = nil
end

local function HideTooltips()
	if not tooltip or MouseIsOver(tooltip) then return end
	tooltip:SetScript("OnUpdate", nil)
	QTip:Release(tooltip)
	tooltip = nil
end

local function ShowIndicatorTooltip(cell, arg, ...)
	local instance = arg.instance
	local toon = arg.toon
	local diff = arg.diff
	if not instance or not toon or not diff then return end
	if indicatortip then QTip:Release(indicatortip) end
	indicatortip = QTip:Acquire("SavedInstancesTooltipChild", 1, "LEFT")
	local hFont = indicatortip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	indicatortip:SetHeaderFont(hFont)
	local thisinstance = SavedInstances.db.Instances[instance][toon]
	indicatortip:AddHeader(strjoin("", NORMALFONT, instance, " (", FONTEND, ClassColorise(SavedInstances.db.Toons[toon].Class, strsplit(' ', toon)), NORMALFONT, ")", FONTEND))
	if thisinstance.Normal then
		indicatortip:AddLine(strjoin("", YELLOWFONT, L["Normal ID"], ": ", FONTEND, thisinstance.Normal.ID))
		if not thisinstance.Heroic or thisinstance.Normal.Expires ~= thisinstance.Heroic.Expires then
			indicatortip:AddLine(strjoin("", YELLOWFONT, L["Time Left"], ": ", FONTEND, SecondsToTime(thisinstance.Normal.Expires - time())))
		end
	end
	if thisinstance.Heroic then
		indicatortip:AddLine(strjoin("", YELLOWFONT, L["Heroic ID"], ": ", FONTEND, thisinstance.Heroic.ID))
		indicatortip:AddLine(strjoin("", YELLOWFONT, L["Time Left"], ": ", FONTEND, SecondsToTime(thisinstance.Heroic.Expires - time())))
	end
	indicatortip:SmartAnchorTo(tooltip)
	indicatortip:Show()
end

local function ShowTooltip(anchorframe)
	if tooltip then QTip:Release(tooltip) end
	tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	CullStaleInstances()
	if columns then columns = nil end
	local columns = { }
	if not SavedInstances.db.Toons[thisToon].Hide and SavedInstances.db.Tooltip.FreshToons then
		columns[thisToon] = columns[thisToon] or tooltip:AddColumn("CENTER")
	end
	local headLine, headCol = tooltip:AddHeader(NORMALFONT .. "SavedInstances" .. FONTEND)
	-- TODO: fix this so it does separate columns and everything nice
	for c, category in ipairs(OrderedCategories()) do
		for i, instance in ipairs(OrderedInstances(category)) do
			local currInstance = SavedInstances.db.Instances[instance]
			if TableLen(currInstance) > 2 then -- instance has toon info
				local currLine, _ = tooltip:AddLine(GRAYFONT .. instance .. FONTEND)
				for toon, v in pairs(SavedInstances.db.Toons) do
					if currInstance[toon] and not SavedInstances.db.Toons[toon].Hide then
						tooltip:SetCell(currLine, 1, NORMALFONT .. instance .. FONTEND)
						if not columns[toon] then
							columns[toon] = tooltip:AddColumn("CENTER")
						end
						local scriptarg = {}
						scriptarg.instance = instance
						scriptarg.toon = toon
						tooltip:SetCell(currLine, columns[toon], DifficultyString(currInstance[toon]))
						tooltip:SetCellScript(currLine, columns[toon], "OnEnter", ShowIndicatorTooltip, scriptarg)
						tooltip:SetCellScript(currLine, columns[toon], "OnLeave", HideIndicatorTooltip)
					end
				end
			elseif SavedInstances.db.Tooltip.AllRaids and currInstance.Show then
				-- instance has no toon info, but must be shown anyway
				tooltip:AddLine(strjoin("", GRAYFONT, instance, FONTEND))
			end
		end
	end
	if tooltip:GetLineCount() == 1 then
		local noneLine = tooltip:AddLine()
		tooltip:SetCell(noneLine, 1, strjoin("", GRAYFONT, NO_RAID_INSTANCES_SAVED, FONTEND), "LEFT", tooltip:GetColumnCount())
	end
	if SavedInstances.db.Tooltip.FreshToons then
		for toon, v in pairs(SavedInstances.db.Toons) do
			if not SavedInstances.db.Toons[toon].Hide then
				columns[toon] = columns[toon] or tooltip:AddColumn("CENTER")
			end
		end
	end
	for toon, col in pairs(columns) do
		tooltip:SetCell(headLine, col, ClassColorise(SavedInstances.db.Toons[toon].Class, strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER")
	end 
	tooltip:AddSeparator(8,0,0,0,0)
	local hintLine, hintCol
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Left-click|r to show Blizzard's Raid Information"], "LEFT", tooltip:GetColumnCount())
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Right-click|r to configure SavedInstances"], "LEFT", tooltip:GetColumnCount())
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["Hover mouse on indicator for details"], "LEFT", tooltip:GetColumnCount())
	tooltip:SmartAnchorTo(anchorframe)
	tooltip:Show()
end

-- global addon code below

function SavedInstances:OnInitialize()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	-- begin backwards compatibility
	if not SavedInstancesDB.DBVersion then
		SavedInstancesDB = defaultDB
	end
	if SavedInstancesDB.DBVersion == 2 then
		SavedInstancesDB.DBVersion = 3
		SavedInstancesDB.Tooltip.NewInstanceShow = false
		SavedInstancesDB.FuBar = nil
		SavedInstancesDB.profile = nil
	end
	if SavedInstancesDB.DBVersion == 3 then
		SavedInstancesDB.DBVersion = 4
		SavedInstancesDB.Indicators = defaultDB.Indicators
	end
	if SavedInstancesDB.DBVersion == 4 then
		SavedInstancesDB.DBVersion = 5
		SavedInstancesDB.Indicators.UnsavedIndicator = nil
		SavedInstancesDB.Indicators.UnsavedText = nil
		SavedInstancesDB.Indicators.UnsavedColor = nil
		for toon, a in pairs(SavedInstancesDB.Toons) do
			a.Class = strupper(gsub(a.Class, " ", ""))
			a.NumSaved = nil
		end
	end
	if SavedInstancesDB.DBVersion == 5 then
		SavedInstancesDB.DBVersion = 6
		SavedInstancesDB.Indicators = defaultDB.Indicators
		SavedInstancesDB.Categories = defaultDB.Categories
		SavedInstancesDB.Tooltip.ShowCategories = true
		SavedInstancesDB.Tooltip.ShowSoloCategory = false
		for instance, a in pairs(SavedInstancesDB.Instances) do
			for toon, b in pairs(SavedInstancesDB.Toons) do
				if a[toon] then
					if a[toon]["Normal"] then
						a[toon][1] = a[toon]["Normal"]
						a[toon]["Normal"] = nil
					end
					if a[toon]["Heroic"] then
						a[toon][2] = a[toon]["Heroic"]
						a[toon]["Heroic"] = nil
					end
				end
			end
		end
	end
	if SavedInstancesDB.DBVersion ~= 6 then
		SavedInstancesDB = defaultDB
	end
	-- end backwards compatibilty
	self.db = db or SavedInstancesDB
	self.defaultDB = defaultDB
	self.db.Toons[thisToon] = self.db.Toons[thisToon] or { }
	self.db.Toons[thisToon].Class = select(2, UnitClass("player"))
	self.db.Toons[thisToon].Hide = self.db.Toons[thisToon].Hide or false
	self:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(frame, button)
				if button == "LeftButton" then
					ToggleFriendsFrame(5) -- open Blizzard Raid window
					RaidInfoFrame:Show()
				else
					SavedInstances:ShowConfig()
				end
			end,
			OnEnter = function(self)
				ShowTooltip(self)
			end,
			OnLeave = function(self)
				if tooltip and MouseIsOver(tooltip) then 
					tooltip:SetScript("OnUpdate", HideTooltips)
				else
					HideTooltips()
				end
			end,
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	CullStaleInstances()
	local numsaved = GetNumSavedInstances()
	if numsaved > 0 then
		for i = 1, numsaved do
			local name, id, expires, diff, locked, extended, mostsig, raid = GetSavedInstanceInfo(i)
			self.db.Instances[name] = self.db.Instances[name] or { }
			self.db.Instances[name].Raid = self.db.Instances[name].Raid or raid or nil
			self.db.Instances[name].Order = self.db.Instances[name].Order or CategorySize(InstanceCategory(name))
			self.db.Instances[name].Show = self.db.Instances[name].Show or self.db.Tooltip.NewInstanceShow
			self.db.Instances[name][thisToon] = self.db.Instances[name][thisToon] or { }
			self.db.Instances[name][thisToon][diff] = {
				ID = id,
				Expires = expires + time()
			}
		end
	end
end

-- options functions below

local function IndicatorIconOptions(nextorder)
	return { order = nextorder, type = "select", width = "half", name = EMBLEM_SYMBOL, values = Indicators }
end

local function IndicatorTextOptions(nextorder)
	return { order = nextorder, type = "input", name = L["Text"], multiline = false }
end

local function IndicatorColorOptions(nextorder)
	return { order = nextorder, type = "color", width = "half", hasAlpha = false, name = COLOR,
		get = function(info)
			local r = SavedInstances.db.Indicators[info[#info]][1]
			local g = SavedInstances.db.Indicators[info[#info]][2]
			local b = SavedInstances.db.Indicators[info[#info]][3]
			return r, g, b, nil
		end,
		set = function(info, r, g, b, ...)
			SavedInstances.db.Indicators[info[#info]][1] = r
			SavedInstances.db.Indicators[info[#info]][2] = g
			SavedInstances.db.Indicators[info[#info]][3] = b
		end,
	}
end
					
-- options table below

SavedInstances.Options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			get = function(info)
					return SavedInstances.db.Tooltip[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Tooltip[info[#info]] = value
			end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapIcon = {
					type = "toggle",
					name = L["Show minimap button"],
					desc = L["Show the SavedInstances minimap button"],
					order = 2,
					hidden = function() return not LDBIcon end,
					get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
					set = function(info, value)
						SavedInstances.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
					end,
				},
				CategoriesHeader = {
					order = 3, 
					type = "header",
					name = L["Categories"],
				},
				ShowCategories = {
					type = "toggle",
					name = L["Show category names"],
					desc = L["Show category names in the tooltip"],
					order = 4,
				},
				ShowSoloCategory = {
					type = "toggle",
					name = L["Single category name"],
					desc = L["Show name for a category when all displayed instances belong only to that category"],
					order = 5,
				},
				InstancesHeader = {
					order = 6, 
					type = "header",
					name = L["Instances"],
				},
				AllRaids = {
					type = "toggle",
					order = 7,
					name = L["Show all instances"],
					desc = L["Always show all instances, even if no characters are saved to them"],
				},
				NewInstanceShow = {
					type = "toggle",
					order = 8,
					name = L["Always show new instances"],
					desc = L["New instances are shown by default, even if no characters are saved to them"],
				},
				InstanceList = {
					order = 9,
					type = "multiselect",
					dialogControl = "Dropdown",
					width = "double",
					name = L["Show these instances when all instances are shown"],
					disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
					get = function(info, instance)
						return SavedInstances.db.Instances[instance].Show
					end,
					set = function(info, instance, value)
						SavedInstances.db.Instances[instance].Show = value
					end,
					values = function()
						local list = { }
						for instance, v in pairs(SavedInstances.db.Instances) do
							list[instance] = instance
						end
						return list
					end,
				},
				ToonHeader = {
					order = 10, 
					type = "header",
					name = L["Characters"],
				},
				FreshToons = {
					type = "toggle",
					name = L["Show non-saved characters"],
					desc = L["Show characters that have not been saved to any instances"],
					order = 11,
				},
				ToonList = {
					order = 12,
					type = "multiselect",
					dialogControl = "Dropdown",
					width = "double",
					name = L["Always hide these characters"],
					get = function(info, toon)
						return SavedInstances.db.Toons[toon].Hide
					end,
					set = function(info, toon, value)
						SavedInstances.db.Toons[toon].Hide = value
					end,
					values = function()
						local list = { }
						for k, v in pairs(SavedInstances.db.Toons) do
							list[k] = k
						end
						return list
					end,
				},
			},
		},
		Indicators = {
			order = 2,
			type = "group",
			name = L["Indicators"],
			get = function(info)
					return SavedInstances.db.Indicators[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Indicators[info[#info]] = value
			end,
			args = {
				Instructions = {
					order = 1,
					type = "description",
					name = L["You can combine icons and text in a single indicator if you wish. Simply choose an icon, and insert the word ICON into the text field. Anywhere the word ICON is found, the icon you chose will be substituted in."],
				},

				D1Header = {
					order = 2, 
					type = "header",
					name = DUNGEON_DIFFICULTY1,
				},
				D1Indicator = IndicatorIconOptions(3),
				D1Text = IndicatorTextOptions(4),
				D1Color = IndicatorColorOptions(5),

				D2Header = {
					order = 6, 
					type = "header",
					name = DUNGEON_DIFFICULTY2,
				},
				D2Indicator = IndicatorIconOptions(7),
				D2Text = IndicatorTextOptions(8),
				D2Color = IndicatorColorOptions(9),
				
				R1Header = {
					order = 10, 
					type = "header",
					name = RAID_DIFFICULTY1,
				},
				R1Indicator = IndicatorIconOptions(11),
				R1Text = IndicatorTextOptions(12),
				R1Color = IndicatorColorOptions(13),

				R2Header = {
					order = 14, 
					type = "header",
					name = RAID_DIFFICULTY2,
				},
				R2Indicator = IndicatorIconOptions(15),
				R2Text = IndicatorTextOptions(16),
				R2Color = IndicatorColorOptions(17),
				
				R3Header = {
					order = 18, 
					type = "header",
					name = RAID_DIFFICULTY3,
				},
				R3Indicator = IndicatorIconOptions(19),
				R3Text = IndicatorTextOptions(20),
				R3Color = IndicatorColorOptions(21),

				R4Header = {
					order = 22, 
					type = "header",
					name = RAID_DIFFICULTY4,
				},
				R4Indicator = IndicatorIconOptions(23),
				R4Text = IndicatorTextOptions(24),
				R4Color = IndicatorColorOptions(25),

				DefaultsHeader = {
					order = -2, 
					type = "header",
					name = DEFAULTS,
				},
				ResetButton = {
					order = -1,
					type = "execute",
					name = RESET_TO_DEFAULT,
					func = function()
						SavedInstances.db.Indicators = SavedInstances.defaultDB.Indicators
					end,
				},
			},
		},
		Categories = {
			order = 3,
			type = "group",
			name = L["Categories"],
			get = function(info)
					return SavedInstances.db.Categories[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Categories[info[#info]] = value
			end,
			args = {
				SelectedCategory = {
					order = 1,
					type = "select",
					name = L["Selected category"],
					get = function(info, category)
						local category = SavedInstances.db.Categories[SavedInstances.selectedCategory]
						if not category then return end
						return format("%02d%s", category.Order, SavedInstances.selectedCategory)
					end,
					set = function(info, category)
						local trimmedcat = strsub(category, 3)
						SavedInstances.selectedCategory = trimmedcat
						SavedInstances.selectedInstance = nil
					end,
					values = function()
						local list = { }
						for k, v in pairs(SavedInstances.db.Categories) do
							list[format("%02d%s", v.Order, k)] = v.Name
						end
						return list
					end,
				},
				
				ReorderHeader = {
					order = 2, 
					type = "header",
					name = L["Category order"],
				},
				Position = {
					order = 3,
					type = "range",
					min = 1,
					max = CategorySize(SavedInstances.selectedCategory),
					step = 1,
					name = L["Position in list"],
					disabled = function()
						return (SavedInstances.selectedCategory == nil)
					end,
					get = function(info, order)
						if not SavedInstances.selectedCategory then return end
						return SavedInstances.db.Categories[SavedInstances.selectedCategory].Order
					end,
					set = function(info, order)
						SavedInstances.db.Categories[SavedInstances.selectedCategory].Order = order
					end,
				},

				DefaultsHeader = {
					order = -2, 
					type = "header",
					name = DEFAULTS,
				},
				ResetButton = {
					order = -1,
					type = "execute",
					name = RESET_TO_DEFAULT,
					func = function()
						SavedInstances.db.Categories = SavedInstances.defaultDB.Categories
					end,
				},
			},
		},
		Instances = {
			order = 4,
			type = "group",
			name = L["Instance options"],
			disabled = function()
				return (SavedInstances.selectedInstance == nil) or (InstanceCategory(SavedInstances.selectedInstance) ~= SavedInstances.selectedCategory)
			end,
			args = {
				SelectedCategory = {
					order = 1,
					type = "select",
					name = L["Selected category"],
					disabled = false,
					get = function(info, category)
						local category = SavedInstances.db.Categories[SavedInstances.selectedCategory]
						if not category then return end
						return format("%02d%s", category.Order, SavedInstances.selectedCategory)
					end,
					set = function(info, category)
						local trimmedcat = strsub(category, 3)
						SavedInstances.selectedCategory = trimmedcat
						SavedInstances.selectedInstance = nil
					end,
					values = function()
						local list = { }
						for k, v in pairs(SavedInstances.db.Categories) do
							list[format("%02d%s", v.Order, k)] = v.Name
						end
						return list
					end,
				},
				SelectedInstance = {
					order = 2,
					type = "select",
					name = L["Selected instance"],
					disabled = function()
						return (SavedInstances.selectedCategory == nil) or
							(TableLen(OrderedInstances(SavedInstances.selectedCategory)) == 0)
					end,
					get = function(info, instance)
						return SavedInstances.selectedInstance
					end,
					set = function(info, instance)
						SavedInstances.selectedInstance = instance
					end,
					values = function()
						local list = { }
						for i, v in ipairs(OrderedInstances(SavedInstances.selectedCategory)) do
							list[v] = v
						end
						return list
					end,
				},

				OrderHeader = {
					order = 3, 
					type = "header",
					name = L["Instance order"],
				},
				Position = {
					order = 4,
					type = "range",
					min = 1,
					max = CategorySize(SavedInstances.selectedCategory),
					step = 1,
					name = L["Position in list"],
					get = function(info, order)
						if not SavedInstances.selectedInstance then return end
						return SavedInstances.db.Instances[SavedInstances.selectedInstance].Order
					end,
					set = function(info, order)
						SavedInstances.db.Instances[SavedInstances.selectedInstance].Order = order
					end,
				},

				DetailsHeader = {
					order = 5, 
					type = "header",
					name = L["Instance details"],
				},
				Expansion = {
					order = 6,
					type = "select",
					name = L["Expansion"],
					get = function(info)
						if SavedInstances.selectedInstance then
							return SavedInstances.db.Instances[SavedInstances.selectedInstance].Expansion
						end
					end,
					set = function(info, expansion)
						local instance = SavedInstances.db.Instances[SavedInstances.selectedInstance]
						instance.Expansion = expansion
						instance.Raid = instance.Raid or false
						SavedInstances.selectedCategory = InstanceCategory(SavedInstances.selectedInstance)
					end,
					values = {
						[0] = EXPANSION_NAME0,
						[1] = EXPANSION_NAME1,
						[2] = EXPANSION_NAME2,
					},
				},
				Raid = {
					order = 7,
					type = "toggle",
					width = "half",
					name = LFG_TYPE_RAID,
					get = function(info)
						if SavedInstances.selectedInstance then
							return SavedInstances.db.Instances[SavedInstances.selectedInstance].Raid
						end
					end,
					set = function(info, value)
						SavedInstances.db.Instances[SavedInstances.selectedInstance].Raid = value
						SavedInstances.selectedCategory = InstanceCategory(SavedInstances.selectedInstance)
					end,
				},
				DeleteButton = {
					order = 8,
					type = "execute",
					width = "half",
					name = DELETE,
					func = function()
						SavedInstances.db.Instances[SavedInstances.selectedInstance] = nil
					end,
				},
			},
		},
	},
}

local optionsFrames = { }
function SavedInstances:SetupOptions()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", SavedInstances.Options)
	optionsFrames["General"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", nil, nil, "General")
	optionsFrames["Indicators"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Indicators"], "SavedInstances", "Indicators")
	optionsFrames["Categories"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Categories"], "SavedInstances", "Categories")
	optionsFrames["Instances"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Instances"], "SavedInstances", "Instances")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["Instances"])
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["General"])
end

